# Copyright 2021 Alethea Katherine Flowers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import contextlib
import sys
from typing import Optional

from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.version import InvalidVersion, Version

if sys.version_info >= (3, 8):  # pragma: no cover
    import importlib.metadata as metadata
else:  # pragma: no cover
    import importlib_metadata as metadata


class VersionCheckFailed(Exception):
    """The Nox version does not satisfy what ``nox.needs_version`` specifies."""


class InvalidVersionSpecifier(Exception):
    """The ``nox.needs_version`` specifier cannot be parsed."""


def get_nox_version() -> str:
    """Return the version of the installed Nox package."""
    return metadata.version("nox")  # type: ignore[no-untyped-call]


def _parse_string_constant(node: ast.AST) -> Optional[str]:  # pragma: no cover
    """Return the value of a string constant."""
    if sys.version_info < (3, 8):
        if isinstance(node, ast.Str) and isinstance(node.s, str):
            return node.s
    elif isinstance(node, ast.Constant) and isinstance(node.value, str):
        return node.value
    return None


def _parse_needs_version(source: str, filename: str = "<unknown>") -> Optional[str]:
    """Parse ``nox.needs_version`` from the user's noxfile."""
    value: Optional[str] = None
    module: ast.Module = ast.parse(source, filename=filename)
    for statement in module.body:
        if isinstance(statement, ast.Assign):
            for target in statement.targets:
                if (
                    isinstance(target, ast.Attribute)
                    and isinstance(target.value, ast.Name)
                    and target.value.id == "nox"
                    and target.attr == "needs_version"
                ):
                    value = _parse_string_constant(statement.value)
    return value


def _read_needs_version(filename: str) -> Optional[str]:
    """Read ``nox.needs_version`` from the user's noxfile."""
    with open(filename) as io:
        source = io.read()

    return _parse_needs_version(source, filename=filename)


def _check_nox_version_satisfies(needs_version: str) -> None:
    """Check if the Nox version satisfies the given specifiers."""
    version = Version(get_nox_version())

    try:
        specifiers = SpecifierSet(needs_version)
    except InvalidSpecifier as error:
        message = f"Cannot parse `nox.needs_version`: {error}"
        with contextlib.suppress(InvalidVersion):
            Version(needs_version)
            message += f", did you mean '>= {needs_version}'?"
        raise InvalidVersionSpecifier(message)

    if not specifiers.contains(version, prereleases=True):
        raise VersionCheckFailed(
            f"The Noxfile requires Nox {specifiers}, you have {version}"
        )


def check_nox_version(filename: str) -> None:
    """Check if ``nox.needs_version`` in the user's noxfile is satisfied.

    Args:

        filename: The location of the user's noxfile. ``nox.needs_version`` is
            read from the noxfile by parsing the AST.

    Raises:
        VersionCheckFailed: The Nox version does not satisfy what
            ``nox.needs_version`` specifies.
        InvalidVersionSpecifier: The ``nox.needs_version`` specifier cannot be
            parsed.
    """
    needs_version = _read_needs_version(filename)

    if needs_version is not None:
        _check_nox_version_satisfies(needs_version)



# Copyright 2017 Alethea Katherine Flowers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import locale
import subprocess
import sys
from typing import IO, Mapping, Optional, Sequence, Tuple, Union


def shutdown_process(proc: subprocess.Popen) -> Tuple[bytes, bytes]:
    """Gracefully shutdown a child process."""

    with contextlib.suppress(subprocess.TimeoutExpired):
        return proc.communicate(timeout=0.3)

    proc.terminate()

    with contextlib.suppress(subprocess.TimeoutExpired):
        return proc.communicate(timeout=0.2)

    proc.kill()

    return proc.communicate()


def decode_output(output: bytes) -> str:
    """Try to decode the given bytes with encodings from the system.

    :param output: output to decode
    :raises UnicodeDecodeError: if all encodings fail
    :return: decoded string
    """
    try:
        return output.decode("utf-8")
    except UnicodeDecodeError:
        second_encoding = locale.getpreferredencoding()
        if second_encoding.casefold() in ("utf8", "utf-8"):
            raise

        return output.decode(second_encoding)


def popen(
    args: Sequence[str],
    env: Optional[Mapping[str, str]] = None,
    silent: bool = False,
    stdout: Optional[Union[int, IO]] = None,
    stderr: Union[int, IO] = subprocess.STDOUT,
) -> Tuple[int, str]:
    if silent and stdout is not None:
        raise ValueError(
            "Can not specify silent and stdout; passing a custom stdout always silences the commands output in Nox's log."
        )

    if silent:
        stdout = subprocess.PIPE

    proc = subprocess.Popen(args, env=env, stdout=stdout, stderr=stderr)

    try:
        out, err = proc.communicate()
        sys.stdout.flush()

    except KeyboardInterrupt:
        out, err = shutdown_process(proc)
        if proc.returncode != 0:
            raise

    return_code = proc.wait()

    return return_code, decode_output(out) if out else ""



# Copyright 2017 Alethea Katherine Flowers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import copy
import functools
from typing import Any, Callable, Optional, TypeVar, Union, overload

from ._decorators import Func
from ._typing import Python

F = TypeVar("F", bound=Callable[..., Any])

_REGISTRY: "collections.OrderedDict[str, Func]" = collections.OrderedDict()


@overload
def session_decorator(__func: F) -> F:
    ...


@overload
def session_decorator(
    __func: None = ...,
    python: Python = ...,
    py: Python = ...,
    reuse_venv: Optional[bool] = ...,
    name: Optional[str] = ...,
    venv_backend: Any = ...,
    venv_params: Any = ...,
) -> Callable[[F], F]:
    ...


def session_decorator(
    func: Optional[F] = None,
    python: Python = None,
    py: Python = None,
    reuse_venv: Optional[bool] = None,
    name: Optional[str] = None,
    venv_backend: Any = None,
    venv_params: Any = None,
) -> Union[F, Callable[[F], F]]:
    """Designate the decorated function as a session."""
    # If `func` is provided, then this is the decorator call with the function
    # being sent as part of the Python syntax (`@nox.session`).
    # If `func` is None, however, then this is a plain function call, and it
    # must return the decorator that ultimately is applied
    # (`@nox.session(...)`).
    #
    # This is what makes the syntax with and without parentheses both work.
    if func is None:
        return functools.partial(
            session_decorator,
            python=python,
            py=py,
            reuse_venv=reuse_venv,
            name=name,
            venv_backend=venv_backend,
            venv_params=venv_params,
        )

    if py is not None and python is not None:
        raise ValueError(
            "The py argument to nox.session is an alias for the python "
            "argument, please only specify one."
        )

    if python is None:
        python = py

    fn = Func(func, python, reuse_venv, name, venv_backend, venv_params)
    _REGISTRY[name or func.__name__] = fn
    return fn


def get() -> "collections.OrderedDict[str, Func]":
    """Return a shallow copy of the registry.

    This ensures that the registry is not accidentally modified by
    calling code that retrieves a registry and mutates it.
    """
    return copy.copy(_REGISTRY)



# Copyright 2017 Alethea Katherine Flowers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Naively converts tox.ini files into noxfile.py files."""

import argparse
import pkgutil
from typing import Any, Iterator

import jinja2
import tox.config

_TEMPLATE = jinja2.Template(
    pkgutil.get_data(__name__, "tox_to_nox.jinja2").decode("utf-8"),  # type: ignore[union-attr]
    extensions=["jinja2.ext.do"],
)


def wrapjoin(seq: Iterator[Any]) -> str:
    return ", ".join([f"'{item}'" for item in seq])


def main() -> None:
    parser = argparse.ArgumentParser(description="Converts toxfiles to noxfiles.")
    parser.add_argument("--output", default="noxfile.py")

    args = parser.parse_args()

    config = tox.config.parseconfig([])
    output = _TEMPLATE.render(config=config, wrapjoin=wrapjoin)

    with open(args.output, "w") as outfile:
        outfile.write(output)



# Copyright 2020 Alethea Katherine Flowers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["TYPE_CHECKING", "ClassVar", "NoReturn", "Python"]

import typing as _typing

try:
    from typing import TYPE_CHECKING
except ImportError:
    try:
        from typing_extensions import TYPE_CHECKING
    except ImportError:
        TYPE_CHECKING = False

try:
    from typing import NoReturn
except ImportError:
    try:
        from typing_extensions import NoReturn
    except ImportError:
        pass


try:
    from typing import ClassVar
except ImportError:
    try:
        from typing_extensions import ClassVar
    except ImportError:
        pass

Python = _typing.Optional[_typing.Union[str, _typing.Sequence[str], bool]]
